# Copyright (c) OpenMMLab. All rights reserved.
import itertools
import math
from typing import Iterator, Optional, Sized

import torch
from torch.utils.data import Sampler

from mmengine.dist import get_dist_info, sync_random_seed
from mmengine.registry import DATA_SAMPLERS
import random

@DATA_SAMPLERS.register_module()
class ModifiedSampler(Sampler):
    """The default data sampler for both distributed and non-distributed
    environment.

    It has several differences from the PyTorch ``DistributedSampler`` as
    below:

    1. This sampler supports non-distributed environment.

    2. The round up behaviors are a little different.

       - If ``round_up=True``, this sampler will add extra samples to make the
         number of samples is evenly divisible by the world size. And
         this behavior is the same as the ``DistributedSampler`` with
         ``drop_last=False``.
       - If ``round_up=False``, this sampler won't remove or add any samples
         while the ``DistributedSampler`` with ``drop_last=True`` will remove
         tail samples.

    Args:
        dataset (Sized): The dataset.
        shuffle (bool): Whether shuffle the dataset or not. Defaults to True.
        seed (int, optional): Random seed used to shuffle the sampler if
            :attr:`shuffle=True`. This number should be identical across all
            processes in the distributed group. Defaults to None.
        round_up (bool): Whether to add extra samples to make the number of
            samples evenly divisible by the world size. Defaults to True.
    """

    def __init__(self,
                 dataset: Sized,
                 patientshuffle: bool = True,
                 seed: Optional[int] = None,
                 round_up: bool = True) -> None:
        rank, world_size = get_dist_info()
        self.rank = rank
        self.world_size = world_size

        self.dataset = dataset
        self.patientshuffle = patientshuffle
        if seed is None:
            seed = sync_random_seed()
        self.seed = seed
        self.epoch = 0
        self.round_up = round_up

        if self.round_up:
            self.num_samples = math.ceil(len(self.dataset) / world_size)
            self.total_size = self.num_samples * self.world_size
        else:
            self.num_samples = math.ceil(
                (len(self.dataset) - rank) / world_size)
            self.total_size = len(self.dataset)

    def __iter__(self) -> Iterator[int]:
        """Iterate the indices."""
        # deterministically shuffle based on epoch and seed
        # shuffle new
        if self.patientshuffle:
            random.seed(self.seed + self.epoch)
            num_images = len(self.dataset)
            groups = {}
            for idx in range(num_images):
                # cat_ids = set(self.dataset.get_cat_ids(idx))
                img_path = self.dataset.get_img_path_ids(idx)
                index = img_path.rfind('/')
                patient_info = img_path[:index]
                if patient_info not in groups:
                    groups[patient_info] = []
                groups[patient_info].append(idx)

            # print(groups)
            new_groups = {}
            boundary = 2
            for key, value in groups.items():
                if len(value) > (boundary+1):
                    random.shuffle(value)
                    if len(value) % boundary == 1:
                        num_keys = len(value) // boundary
                        split_values = [value[i:i + boundary] for i in range(0, len(value) - 3, boundary)]
                        split_values.append(value[-3:])
                    else:
                        num_keys = len(value) // boundary
                        split_values = [value[i:i + boundary] for i in range(0, len(value), boundary)]

                    for i in range(num_keys):
                        new_key = f"{key}_{i + 1}"
                        new_groups[new_key] = split_values[i]
                else:
                    new_groups[key] = value

            items = list(new_groups.items())
            # items = list(groups.items())

            random.shuffle(items)
            indices = []
            indices.extend(indices for key, value in items for indices in value)
        else:
            indices = torch.arange(len(self.dataset)).tolist()

        if self.round_up:
            indices = (
                indices *
                int(self.total_size / len(indices) + 1))[:self.total_size]

        # subsample
        # indices = indices[self.rank:self.total_size:self.world_size]

        # subsample new
        total_samples = len(indices)
        samples_per_process = int(math.ceil(total_samples / self.world_size))
        start_index = self.rank * samples_per_process
        end_index = min((self.rank + 1) * samples_per_process, total_samples)
        indices = indices[start_index:end_index]

        return iter(indices)

    def __len__(self) -> int:
        """The number of samples in this rank."""
        return self.num_samples

    def set_epoch(self, epoch: int) -> None:
        """Sets the epoch for this sampler.

        When :attr:`shuffle=True`, this ensures all replicas use a different
        random ordering for each epoch. Otherwise, the next iteration of this
        sampler will yield the same ordering.

        Args:
            epoch (int): Epoch number.
        """
        self.epoch = epoch