from mmengine.dataset import DefaultSampler
# Copyright (c) OpenMMLab. All rights reserved.
import itertools
import math
import bisect
import copy
import random
from typing import Iterator, Optional, Sized
from collections import defaultdict
import torch
from torch.utils.data import Sampler
from mmengine.dataset.base_dataset import BaseDataset
from mmengine.dist import get_dist_info, sync_random_seed
from mmengine.registry import DATA_SAMPLERS

@DATA_SAMPLERS.register_module()
class NonuniformSampler(Sampler):
    """The default data sampler for both distributed and non-distributed
    environment.

    It has several differences from the PyTorch ``DistributedSampler`` as
    below:

    1. This sampler supports non-distributed environment.

    2. The round up behaviors are a little different.

       - If ``round_up=True``, this sampler will add extra samples to make the
         number of samples is evenly divisible by the world size. And
         this behavior is the same as the ``DistributedSampler`` with
         ``drop_last=False``.
       - If ``round_up=False``, this sampler won't remove or add any samples
         while the ``DistributedSampler`` with ``drop_last=True`` will remove
         tail samples.

    Args:
        dataset (Sized): The dataset.
        shuffle (bool): Whether shuffle the dataset or not. Defaults to True.
        seed (int, optional): Random seed used to shuffle the sampler if
            :attr:`shuffle=True`. This number should be identical across all
            processes in the distributed group. Defaults to None.
        round_up (bool): Whether to add extra samples to make the number of
            samples evenly divisible by the world size. Defaults to True.
    """

    def __init__(self,
                 dataset: Sized,
                 shuffle: bool = True,
                 prob: float = 0.125,
                 seed: Optional[int] = None,
                 round_up: bool = True) -> None:
        rank, world_size = get_dist_info()
        self.rank = rank
        self.world_size = world_size

        self.dataset = dataset
        self.shuffle = shuffle
        self.prob = prob
        if seed is None:
            seed = sync_random_seed()
        self.seed = seed
        self.epoch = 0
        self.round_up = round_up
        # print(self.rank)# 0
        if self.round_up:
            self.num_samples = math.ceil(len(self.dataset) / world_size)
            self.total_size = self.num_samples * self.world_size
            # print(world_size)# 1
            # print(self.total_size)
        else:
            self.num_samples = math.ceil(
                (len(self.dataset) - rank) / world_size)
            self.total_size = len(self.dataset)

    def expand_list(self, original_list: list, target_size: int ) -> list:
        original_size = len(original_list)
        repetitions = target_size // original_size  # Determine how many times to repeat the original list
        remainder = target_size % original_size  # Determine the remaining positions to fill
        expanded_list = original_list * repetitions  # Expand the original list by repeating it
        expanded_list += original_list[:remainder]  # Add the remaining elements
        return expanded_list

    def __iter__(self) -> Iterator[int]:
        """Iterate the indices."""

        # 计算这个类的正样本
        category_freq: defaultdict = defaultdict(float)
        num_images = len(self.dataset)
        positive_samples = []
        suspicious_samples = []
        negative_samples = []
        for idx in range(num_images):
            cat_ids = set(self.dataset.get_cat_ids(idx))
            for cat_id in cat_ids:
                if cat_id == 2:
                    positive_samples.append(idx)
                elif cat_id == 1:
                    suspicious_samples.append(idx)
                else:
                    negative_samples.append(idx)
                category_freq[cat_id] += 1
        target_size = int(self.prob * num_images)

        suspicious_samples = self.expand_list(suspicious_samples, target_size)
        positive_samples = self.expand_list(positive_samples, target_size)
        # negative_samples = negative_samples[:(num_images-len(positive_samples))]

        negative_samples = random.sample(negative_samples, num_images-len(positive_samples)-len(suspicious_samples))
        indices = positive_samples + negative_samples + suspicious_samples
        random.shuffle(indices)

        # add extra samples to make it evenly divisible
        if self.round_up:
            indices = (
                indices *
                int(self.total_size / len(indices) + 1))[:self.total_size]
        # subsample
        indices = indices[self.rank:self.total_size:self.world_size]

        return iter(indices)


    def __len__(self) -> int:
        """The number of samples in this rank."""
        return self.num_samples

    def set_epoch(self, epoch: int) -> None:
        """Sets the epoch for this sampler.

        When :attr:`shuffle=True`, this ensures all replicas use a different
        random ordering for each epoch. Otherwise, the next iteration of this
        sampler will yield the same ordering.

        Args:
            epoch (int): Epoch number.
        """
        self.epoch = epoch

