import os
from collections.abc import Sequence
from typing import Optional, Tuple, Union
from numbers import Number
import math
import numpy as np
import mmcv
import mmengine.fileio as fileio
from mmcv.transforms.loading import LoadImageFromFile
from mmcls.registry import TRANSFORMS
import torch
from torch.nn import Upsample
from skimage.transform import rescale
from mmcls.structures import ClsDataSample,MultiTaskDataSample
from .formatting import PackClsInputs, to_tensor

import cv2
from mmengine.logging import MMLogger
from mmengine.dist import get_rank
from mmcv.transforms import BaseTransform
from mmcv.transforms.utils import cache_randomness
from collections import defaultdict
from functools import partial


@TRANSFORMS.register_module()
class LoadOCTImageFromFile(LoadImageFromFile):
    def __init__(self,
                 oct_extensions=('.jpg', '.jpeg', '.png', '.ppm', '.bmp', '.pgm', '.tif'),
                 *args, **kwargs):
        super(LoadOCTImageFromFile, self).__init__(*args, **kwargs)
        self.oct_extensions = oct_extensions

    # Based on mmcv.transforms.loading.LoadImageFromFile.transform
    def _load_img_file(self, filename, color_type):
        try:
            if self.file_client_args is not None:
                file_client = fileio.FileClient.infer_client(
                    self.file_client_args, filename)
                img_bytes = file_client.get(filename)
            else:
                img_bytes = fileio.get(
                    filename, backend_args=self.backend_args)
            img = mmcv.imfrombytes(
                img_bytes, flag=color_type, backend=self.imdecode_backend)
        except Exception as e:
            if self.ignore_empty:
                return None
            else:
                raise e
        if self.to_float32:
            img = img.astype(np.float32)
        return img

    def transform(self, results: dict) -> Optional[dict]:
        # load fundus image
        # XXXXXXXX

        # load oct images
        oct_img_list = []
        oct_img_path = results['img_path']

        if 'Macular' in oct_img_path:
            oct_start = 'volume_1'
        else:
            oct_start = 'volume_0'
        for file in os.listdir(oct_img_path):
            if not file.endswith(self.oct_extensions):
                continue
            if file.startswith(oct_start):
                filename = os.path.join(oct_img_path, file)
                img = self._load_img_file(filename, color_type='grayscale')
                oct_img_list.append(img)

        oct_img = np.stack(oct_img_list, -1)

        # preprocessing based on channel
        if oct_img.shape[2] == 128:
            # downsample to [64,64,128]
            # downsampled_img = oct_img[::16, ::8, :]
            # downsampled_img = rescale(oct_img, (64 / 1024, 64 / 512, 1.0))
            # oct_img = rescale(downsampled_img, (1.0, 1.0, 200 / 128))
            oct_img = rescale(oct_img, (64 / 1024, 64 / 512, 200 / 128))

            # upsample to[64,64,200]
            # output_shape = (64, 64, 200)
            # downsampled_img = torch.from_numpy(downsampled_img)
            # Upsampler = Upsample(size=(64, 64, 200), mode='bilinear', align_corners=True)
            # oct_img = Upsampler(downsampled_img)
            # oct_img = oct_img.detach().numpy()
            # upsampled_img = np.zeros((64, 64, 200))
            # upsampled_img[:, :, :128] = downsampled_img
            # oct_img = upsampled_img.repeat(int(200 / 128), axis=2)

        elif oct_img.shape[2] == 200:
            # down sample to [64,50,200]
            # down sampled_img = oct_img[::16, ::4, :]
            # padding to [64,64,200]
            # pad = 7
            # padded_img = np.zeros((64, 64, 200))
            # padded_img[:, pad:-pad, :] = downsampled_img
            # oct_img = padded_img
            downsampled_img = rescale(oct_img, (64 / 1024, 50 / 200, 1.0))
            oct_img = np.pad(downsampled_img, ((0, 0), (7, 7), (0, 0)), mode='constant')

        results['img'] = oct_img
        results['img_shape'] = oct_img.shape[:2]
        results['ori_shape'] = oct_img.shape[:2]
        return results


@TRANSFORMS.register_module()
class PackOCTInputs(PackClsInputs):
    """Pack the inputs data for the classification.

    **Required Keys:**

    - img
    - gt_label (optional)
    - ``*meta_keys`` (optional)

    **Deleted Keys:**

    All keys in the dict.

    **Added Keys:**

    - inputs (:obj:`torch.Tensor`): The forward data of models.
    - data_samples (:obj:`~mmcls.structures.ClsDataSample`): The annotation
      info of the sample.

    Args:
        meta_keys (Sequence[str]): The meta keys to be saved in the
            ``metainfo`` of the packed ``data_samples``.
            Defaults to a tuple includes keys:

            - ``sample_idx``: The id of the image sample.
            - ``img_path``: The path to the image file.
            - ``ori_shape``: The original shape of the image as a tuple (H, W).
            - ``img_shape``: The shape of the image after the pipeline as a
              tuple (H, W).
            - ``scale_factor``: The scale factor between the resized image and
              the original image.
            - ``flip``: A boolean indicating if image flip transform was used.
            - ``flip_direction``: The flipping direction.
    """

    def __init__(self, meta_keys=('sample_idx', 'img_path', 'ori_shape', 'img_shape',
                                  'scale_factor', 'flip', 'flip_direction'),
                 oct_meta_keys=('oct_img_path', 'oct_ori_shape', 'oct_img_shape',
                                'oct_scale_factor', 'oct_flip', 'oct_flip_direction')):
        super().__init__(meta_keys)
        if oct_meta_keys is not None:
            self.meta_keys += oct_meta_keys

    def transform(self, results: dict) -> dict:
        packed_results = dict()

        # prepare fundus tensor inputs
        # if 'img' in results:
        #     img = results['img']
        #     if len(img.shape) < 3:
        #         img = np.expand_dims(img, -1)
        #     img = np.ascontiguousarray(img.transpose(2, 0, 1))
        #     packed_results['inputs'] = to_tensor(img)

        # NEW
        # prepare oct tensor inputs
        if 'img' in results:
            oct_img = results['img']
            if len(oct_img.shape) < 3:
                oct_img = np.expand_dims(oct_img, -1)
            oct_img = np.ascontiguousarray(oct_img.transpose(2, 0, 1))
            # if 3D resnet
            # oct_img = np.expand_dims(oct_img, axis=2)
            packed_results['inputs'] = to_tensor(oct_img)
        # END NEW

        data_sample = ClsDataSample()
        if 'gt_label' in results:
            gt_label = results['gt_label']
            data_sample.set_gt_label(gt_label)

        img_meta = {k: results[k] for k in self.meta_keys if k in results}
        data_sample.set_metainfo(img_meta)
        packed_results['data_samples'] = data_sample
        return packed_results


@TRANSFORMS.register_module()
class SaveDebugImage:
    def __init__(self, prefix='step0', output_root='./work_debug'):
        """
        Args:
            prefix: 可以是任意数字或字符串，指定调用当前 SaveDebugImage 时保存的文件夹名称
        """
        self.prefix = prefix

        self.output_path = os.path.join(output_root, f'{prefix}')
        os.makedirs(self.output_path, exist_ok=True)

        self.debug_index = 0

    def __call__(self, results):
        if get_rank() != 0:
            # only save debug results on gpu 0
            return results

        if 'img' in results:
            # print keys of transform parameters
            logger = MMLogger.get_current_instance()
            print_results = dict()
            for k, v in results.items():
                if k == 'img' or 'oct_' in k:
                    continue
                print_results[k] = v
            logger.info('[{}]  prefix: {}  {}'.format(self.debug_index, self.prefix, print_results))

            # save results
            img = results['img']
            filename = os.path.basename(results['img_path'])

            output_path = os.path.join(self.output_path, f'{self.debug_index}_{filename}')
            cv2.imwrite(output_path, img)

        self.debug_index += 1
        return results

    def __repr__(self):
        return self.__class__.__name__ + f'(keys={self.keys})'


@TRANSFORMS.register_module()
class SaveDebugOctImage(SaveDebugImage):
    def __call__(self, results):
        if get_rank() != 0:
            # only save debug results on gpu 0
            return results

        # if 'oct_img' in results:
        if 'img' in results:
            # print keys of transform parameters
            logger = MMLogger.get_current_instance()
            print_results = dict()
            for k, v in results.items():
                # if k == 'oct_img' or 'oct_' not in k:
                if k == 'img' or 'oct_' in k:
                    continue
                print_results[k] = v
            logger.info('[{}]  prefix: {}  {}'.format(self.debug_index, self.prefix, print_results))

            # save results
            # oct_img = results['oct_img']
            # filedir = os.path.basename(results['oct_img_path'])
            oct_img = results['img']
            filedir = os.path.basename(results['img_path'])
            output_path = os.path.join(self.output_path, f'{self.debug_index}_{filedir}')
            os.makedirs(output_path, exist_ok=True)

            oct_img = oct_img.transpose(2, 0, 1)  # (H, W, C) -> (C, H, W)
            for index, img in enumerate(oct_img):
                # img = img * 255
                img = img.astype(int)
                cv2.imwrite(os.path.join(output_path, f'{index}.png'), img)

            self.debug_index += 1
        return results


@TRANSFORMS.register_module()
class LoadPtFromFile(LoadImageFromFile):
    def __init__(self, size="200_128_128",
                 *args, **kwargs):
        super(LoadImageFromFile, self).__init__(*args, **kwargs)
        self.size = size
        self.file = 'OCT_200_128_128.pt'

    def transform(self, results: dict) -> Optional[dict]:
        oct_img_path = results['img_path']
        if self.size == "128_128_128" and "Macular" in oct_img_path:
            # self.file = 'OCT_200_128_128.pt'
            # filename = os.path.join(oct_img_path, self.file)
            # oct_img_load = torch.load(filename)
            # oct_img = rescale(oct_img_load, (1, 1, 128 / 200))

            # new
            self.file = 'OCT_128_128_128.pt'
            filename = os.path.join(oct_img_path, self.file)
            oct_img = torch.load(filename)
        elif self.size == "200_64_64":
            self.file = 'OCT_saved.pt'
            filename = os.path.join(oct_img_path, self.file)
            oct_img = torch.load(filename)
        else:
            self.file = 'OCT_200_128_128.pt'
            filename = os.path.join(oct_img_path, self.file)
            oct_img = torch.load(filename)

        # filename = os.path.join(oct_img_path, self.file)
        # oct_img = torch.load(filename)

        results['img'] = oct_img
        results['img_shape'] = oct_img.shape
        results['ori_shape'] = oct_img.shape
        return results


@TRANSFORMS.register_module()
class PADDING(BaseTransform):
    def __init__(self,
                 padding=None,
                 prob=1,
                 pad_val=0,
                 tragert_size=None,
                 padding_mode='constant'):
        # check padding mode
        assert padding_mode in ['constant', 'edge', 'reflect', 'symmetric']
        self.padding = padding
        self.pad_val = pad_val
        self.tragert_size = tragert_size
        self.padding_mode = padding_mode
        self.prob = prob

    @cache_randomness
    def random_disable(self):
        """Randomly disable the transform."""
        return np.random.rand() > self.prob

    def transform(self, results: dict) -> Optional[dict]:
        # new
        if self.random_disable():
            return results
        # end new

        img = results['img']
        if self.tragert_size is not None:
            if img.shape == self.tragert_size:
                return results
        img = mmcv.impad(img, padding=self.padding, pad_val=self.pad_val, padding_mode=self.padding_mode)

        results['img'] = img
        results['img_shape'] = img.shape
        results['ori_shape'] = img.shape

        return results


@TRANSFORMS.register_module()
class NEWRandomCrop(BaseTransform):
    """Crop the given Image at a random location.

    **Required Keys:**

    - img

    **Modified Keys:**

    - img
    - img_shape

    Args:
        crop_size (int | Sequence): Desired output size of the crop. If
            crop_size is an int instead of sequence like (h, w), a square crop
            (crop_size, crop_size) is made.
        padding (int | Sequence, optional): Optional padding on each border
            of the image. If a sequence of length 4 is provided, it is used to
            pad left, top, right, bottom borders respectively.  If a sequence
            of length 2 is provided, it is used to pad left/right, top/bottom
            borders, respectively. Default: None, which means no padding.
        pad_if_needed (bool): It will pad the image if smaller than the
            desired size to avoid raising an exception. Since cropping is done
            after padding, the padding seems to be done at a random offset.
            Default: False.
        pad_val (Number | Sequence[Number]): Pixel pad_val value for constant
            fill. If a tuple of length 3, it is used to pad_val R, G, B
            channels respectively. Default: 0.
        padding_mode (str): Type of padding. Defaults to "constant". Should
            be one of the following:

            - ``constant``: Pads with a constant value, this value is specified
              with pad_val.
            - ``edge``: pads with the last value at the edge of the image.
            - ``reflect``: Pads with reflection of image without repeating the
              last value on the edge. For example, padding [1, 2, 3, 4]
              with 2 elements on both sides in reflect mode will result
              in [3, 2, 1, 2, 3, 4, 3, 2].
            - ``symmetric``: Pads with reflection of image repeating the last
              value on the edge. For example, padding [1, 2, 3, 4] with
              2 elements on both sides in symmetric mode will result in
              [2, 1, 1, 2, 3, 4, 4, 3].
    """

    def __init__(self,
                 crop_size=(128, 128),
                 prob=1,
                 tragert_size=None,
                 padding=None,
                 pad_if_needed=False,
                 pad_val=0,
                 padding_mode='constant',
                 interpolation='bilinear',
                 backend='cv2'
                 ):
        if isinstance(crop_size, Sequence):
            assert len(crop_size) == 2
            assert crop_size[0] > 0 and crop_size[1] > 0
            self.crop_size = crop_size
        else:
            assert crop_size > 0
            self.crop_size = (crop_size, crop_size)
        # check padding mode
        assert padding_mode in ['constant', 'edge', 'reflect', 'symmetric']
        self.padding = padding
        self.pad_if_needed = pad_if_needed
        self.pad_val = pad_val
        self.padding_mode = padding_mode

        self.prob = prob
        self.tragert_size = tragert_size
        self.interpolation = interpolation
        self.backend = backend

    @cache_randomness
    def random_disable(self):
        """Randomly disable the transform."""
        return np.random.rand() > self.prob

    @cache_randomness
    def rand_crop_params(self, img: np.ndarray):
        """Get parameters for ``crop`` for a random crop.

        Args:
            img (ndarray): Image to be cropped.

        Returns:
            tuple: Params (offset_h, offset_w, target_h, target_w) to be
                passed to ``crop`` for random crop.
        """
        h, w = img.shape[:2]
        target_h, target_w = self.crop_size
        if w == target_w and h == target_h:
            return 0, 0, h, w
        elif w < target_w or h < target_h:
            target_w = min(w, target_w)
            target_h = min(w, target_h)

        offset_h = np.random.randint(0, h - target_h + 1)
        offset_w = np.random.randint(0, w - target_w + 1)

        return offset_h, offset_w, target_h, target_w

    def transform(self, results: dict) -> dict:
        """Transform function to randomly crop images.

        Args:
            results (dict): Result dict from loading pipeline.

        Returns:
            dict: Randomly cropped results, 'img_shape'
                key in result dict is updated according to crop size.
        """

        # new
        if self.random_disable():
            return results
        # end new

        img = results['img']
        if self.tragert_size is not None:
            if img.shape == self.tragert_size:
                return results

        if self.padding is not None:
            img = mmcv.impad(img, padding=self.padding, pad_val=self.pad_val)

        # pad img if needed
        if self.pad_if_needed:
            h_pad = math.ceil(max(0, self.crop_size[0] - img.shape[0]) / 2)
            w_pad = math.ceil(max(0, self.crop_size[1] - img.shape[1]) / 2)

            img = mmcv.impad(
                img,
                padding=(w_pad, h_pad, w_pad, h_pad),
                pad_val=self.pad_val,
                padding_mode=self.padding_mode)

        offset_h, offset_w, target_h, target_w = self.rand_crop_params(img)
        img = mmcv.imcrop(
            img,
            np.array([
                offset_w,
                offset_h,
                offset_w + target_w - 1,
                offset_h + target_h - 1,
            ]))
        results['img'] = img
        results['img_shape'] = img.shape
        results['ori_shape'] = img.shape

        return results

    def __repr__(self):
        """Print the basic information of the transform.

        Returns:
            str: Formatted string.
        """
        repr_str = self.__class__.__name__ + f'(crop_size={self.crop_size}'
        repr_str += f', padding={self.padding}'
        repr_str += f', pad_if_needed={self.pad_if_needed}'
        repr_str += f', pad_val={self.pad_val}'
        repr_str += f', padding_mode={self.padding_mode})'
        return repr_str


@TRANSFORMS.register_module()
class SliceAug(BaseTransform):
    def __init__(self, chan=200, *args, **kwargs):
        super(BaseTransform, self).__init__(*args, **kwargs)
        self.chan = chan

    def transform(self, results: dict) -> Optional[dict]:

        img = results['img']
        TF = np.random.randint(0, 3)
        a = np.random.randint(1, int(self.chan*0.32))  # if chan = 200, max_a =64, if chan = 128 max_a = 40

        if TF == 0:
            return results
        elif TF == 1:
            # CUT Front
            selected_slices = img[:, :, a:self.chan-1]
            duplicated_slice = np.tile(img[:, :, self.chan-1], (a + 1, 1, 1))
            duplicated_slice = np.transpose(duplicated_slice, (1, 2, 0))
            augmented_image = np.concatenate((selected_slices, duplicated_slice), axis=2)
            results['img'] = augmented_image
            results['img_shape'] = augmented_image.shape
            results['ori_shape'] = augmented_image.shape
            return results
        elif TF == 2:
            # CUT Tail
            selected_slices = img[:, :, 0:-a]
            duplicated_slice = np.tile(img[:, :, 0], (a, 1, 1))
            duplicated_slice = np.transpose(duplicated_slice, (1, 2, 0))
            augmented_image = np.concatenate((duplicated_slice, selected_slices), axis=2)
            results['img'] = augmented_image
            results['img_shape'] = augmented_image.shape
            results['ori_shape'] = augmented_image.shape
            return results


@TRANSFORMS.register_module()
class RandomRotate(BaseTransform):
    """Rotate the image.
    Args:
        prob (float): The rotation probability.
        degree (float, tuple[float]): Range of degrees to select from. If
            degree is a number instead of tuple like (min, max),
            the range of degree will be (``-degree``, ``+degree``)
        pad_val (float, optional): Padding value of image. Default: 0.
        center (tuple[float], optional): Center point (w, h) of the rotation in
            the source image. If not specified, the center of the image will be
            used. Default: None.
        auto_bound (bool): Whether to adjust the image size to cover the whole
            rotated image. Default: False
    """

    def __init__(self,
                 prob,
                 degree,
                 pad_val=0,
                 center=None,
                 auto_bound=False):
        self.prob = prob
        assert prob >= 0 and prob <= 1
        if isinstance(degree, (float, int)):
            assert degree > 0, f'degree {degree} should be positive'
            self.degree = (-degree, degree)
        else:
            self.degree = degree
        assert len(self.degree) == 2, f'degree {self.degree} should be a ' \
                                      f'tuple of (min, max)'
        self.pal_val = pad_val
        self.center = center
        self.auto_bound = auto_bound

    def transform(self, results):
        rotate = True if np.random.rand() < self.prob else False
        degree = np.random.uniform(min(*self.degree), max(*self.degree))
        if rotate:
            # rotate image
            results['img'] = mmcv.imrotate(
                results['img'],
                angle=degree,
                border_value=self.pal_val,
                center=self.center,
                auto_bound=self.auto_bound)
        return results

    def __repr__(self):
        repr_str = self.__class__.__name__
        repr_str += f'(prob={self.prob}, ' \
                    f'degree={self.degree}, ' \
                    f'pad_val={self.pal_val}, ' \
                    f'center={self.center}, ' \
                    f'auto_bound={self.auto_bound})'
        return repr_str

@TRANSFORMS.register_module()
class Modified_PackMultiTaskInputs(BaseTransform):
    """Convert all image labels of multi-task dataset to a dict of tensor.

    Args:
        tasks (List[str]): The task names defined in the dataset.
        meta_keys(Sequence[str]): The meta keys to be saved in the
            ``metainfo`` of the packed ``data_samples``.
            Defaults to a tuple includes keys:

            - ``sample_idx``: The id of the image sample.
            - ``img_path``: The path to the image file.
            - ``ori_shape``: The original shape of the image as a tuple (H, W).
            - ``img_shape``: The shape of the image after the pipeline as a
              tuple (H, W).
            - ``scale_factor``: The scale factor between the resized image and
              the original image.
            - ``flip``: A boolean indicating if image flip transform was used.
            - ``flip_direction``: The flipping direction.
    """

    def __init__(self,
                 task_handlers=dict(),
                 multi_task_fields=('gt_label', ),
                 meta_keys=('sample_idx', 'img_path', 'ori_shape', 'img_shape',
                            'scale_factor', 'flip', 'flip_direction')):
        self.multi_task_fields = multi_task_fields
        self.meta_keys = meta_keys
        self.task_handlers = defaultdict(
            partial(PackClsInputs, meta_keys=meta_keys))
        for task_name, task_handler in task_handlers.items():
            self.task_handlers[task_name] = TRANSFORMS.build(
                dict(type=task_handler, meta_keys=meta_keys))

    def transform(self, results: dict) -> dict:
        """Method to pack the input data.

        result = {'img_path': 'a.png', 'gt_label': {'task1': 1, 'task3': 3},
            'img': array([[[  0,   0,   0])
        """
        packed_results = dict()
        results = results.copy()

        if 'img' in results:
            img = results.pop('img')
            if len(img.shape) < 3:
                img = np.expand_dims(img, -1)
            img = np.ascontiguousarray(img.transpose(2, 0, 1))
            # new
            img = np.expand_dims(img, axis=0)
            # end new
            packed_results['inputs'] = to_tensor(img)

        task_results = defaultdict(dict)
        for field in self.multi_task_fields:
            if field in results:
                value = results.pop(field)
                for k, v in value.items():
                    task_results[k].update({field: v})

        data_sample = MultiTaskDataSample()
        for task_name, task_result in task_results.items():
            task_handler = self.task_handlers[task_name]
            task_pack_result = task_handler({**results, **task_result})
            data_sample.set_field(task_pack_result['data_samples'], task_name)

        packed_results['data_samples'] = data_sample
        return packed_results

    def __repr__(self):
        repr = self.__class__.__name__
        task_handlers = {
            name: handler.__class__.__name__
            for name, handler in self.task_handlers.items()
        }
        repr += f'(task_handlers={task_handlers}, '
        repr += f'multi_task_fields={self.multi_task_fields}, '
        repr += f'meta_keys={self.meta_keys})'
        return repr