# Copyright (c) OpenMMLab. All rights reserved.
import torch
import torch.nn as nn

from mmcls.registry import MODELS


@MODELS.register_module()
class ModifiedGlobalAveragePooling(nn.Module):
    """Global Average Pooling neck.

    Note that we use `view` to remove extra channel after pooling. We do not
    use `squeeze` as it will also remove the batch dimension when the tensor
    has a batch dimension of size 1, which can lead to unexpected errors.

    Args:
        dim (int): Dimensions of each sample channel, can be one of {1, 2, 3}.
            Default: 2
    """

    def __init__(self, dim=2, TaskNum=3, Require_fuse=True):
        super(ModifiedGlobalAveragePooling, self).__init__()
        assert dim in [1, 2, 3], 'GlobalAveragePooling dim only support ' \
                                 f'{1, 2, 3}, get {dim} instead.'
        if dim == 1:
            self.gap = nn.AdaptiveAvgPool1d(1)
        elif dim == 2:
            self.gap = nn.AdaptiveAvgPool2d((1, 1))
        else:
            self.gap = nn.AdaptiveAvgPool3d((1, 1, 1))
        self.taskNum = TaskNum
        self.gap1d = nn.AdaptiveAvgPool1d(1)
        self.require_fuse = Require_fuse

    def init_weights(self):
        pass

    def forward(self, inputs):
        if isinstance(inputs, tuple):
            # outs = tuple([self.gap(x) for x in inputs])
            # outs = tuple(
            #     [out.view(x.size(0), -1) for out, x in zip(outs, inputs)])
            x = inputs[0]
            Task_specific_features = inputs[1]
            outs = self.gap(x)
            outs = outs.view(x.size(0), -1)

            Fused_specific_features = []
            if self.require_fuse:
                for i in range(self.taskNum):
                    Task_features = []
                    for stage in Task_specific_features:
                        # fused_block = torch.zeros(3)
                        # fused_block = torch.mean(stage, dim=0)
                        # Fused_specific_features.append(fused_block)
                        # feature = torch.zeros(stage[0][0].size())
                        feature = stage[0][i].clone()
                        for block in stage[1:]:
                            feature += block[i]
                        feature = feature / len(stage)

                        Task_features.append(feature)

                    Fused_specific_features.append(Task_features)
            else:
                for i in range(self.taskNum):
                    Task_features = []
                    for stage in Task_specific_features:
                        feature = stage[i]
                        Task_features.append(feature)
                    Fused_specific_features.append(Task_features)

            Gap_Task_specific_features = []
            for i in range(self.taskNum):
                Task_features = []
                task_feature = Fused_specific_features[i]
                for stage_feature in task_feature:
                    stage_feature = self.gap1d(stage_feature.permute(0, 2, 1))
                    stage_feature = stage_feature.view(stage_feature.size(0), -1)
                    Task_features.append(stage_feature)

                concatenated_Task_tensor = torch.cat(Task_features, dim=1)
                Gap_Task_specific_features.append(concatenated_Task_tensor)

            Gap_Task_specific_features.append(outs)
            outs = tuple(Gap_Task_specific_features)

        elif isinstance(inputs, torch.Tensor):
            outs = self.gap(inputs)
            outs = outs.view(inputs.size(0), -1)
        else:
            raise TypeError('neck inputs should be tuple or torch.tensor')
        return outs
